*DECK SPOASM
      SUBROUTINE SPOASM(IPSYS,IPMACR,IPTRK,IPSNAP,IMPX)
*
*-----------------------------------------------------------------------
*
*Purpose:
* Dragon assembly of SPOT information.
*
*Copyright:
* Copyright (C) 2024 Ecole Polytechnique de Montreal
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version
*
*Author(s): A. Hebert
*
*Parameters: input
* IPSYS   pointer to the pij LCM object.
* IPMACR  pointer to the macrolib LCM object.
* IPTRK   pointer to the tracking LCM object.
* IPSNAP  pointer to the snapshot archive object.
* IMPX    print flag (equal to zero for no print).
*
*-----------------------------------------------------------------------
*
      USE GANLIB
*----
*  SUBROUTINE ARGUMENTS
*----
      TYPE(C_PTR) IPSYS,IPMACR,IPTRK,IPSNAP
      INTEGER IMPX
*----
*  LOCAL VARIABLES
*----
      PARAMETER(NSTATE=40,IOUT=6)
      INTEGER IPAR(NSTATE),ISTATE(NSTATE)
      DOUBLE PRECISION SSUM
      TYPE(C_PTR) JPSNAP1,JPSNAP2,JPSNAP3,KPSNAP,LPSNAP,MPSNAP,IPFLUX,
     1 IPSOUR,JPSYS,KPSYS
      CHARACTER HSMG*131
*----
*  ALLOCATABLE ARRAYS
*----
      INTEGER, ALLOCATABLE, DIMENSION(:) :: IDL2D,MAT2D
      REAL, ALLOCATABLE, DIMENSION(:) :: TXSC2D,LEAK1D
      REAL, ALLOCATABLE, DIMENSION(:,:) :: SXSC2D,TXSC
      REAL, ALLOCATABLE, DIMENSION(:,:,:) :: SXSC,FLUX2D,SOUR2D
      DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: W
      DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: U,V
      DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:,:) :: DB2
*----
*  RECOVER GENERAL TRACKING INFORMATION
*----
      CALL LCMGET(IPTRK,'STATE-VECTOR',IPAR)
      NREG=IPAR(1)
      NUNKN=IPAR(2)
      NMIX=IPAR(4)
      NREG2D=IPAR(6)
      NFLOOR=IPAR(7)
      ISCAT=IPAR(16)
*----
*  RECOVER 2D TRACKING INFORMATION FROM IPSNAP
*----
      ALLOCATE(MAT2D(NREG),IDL2D(NREG))
      CALL LCMGET(IPSNAP,'LISTDIM',NSNAP)
      JPSNAP1=LCMGID(IPSNAP,'TRACK')
      NUNK2D=0
      NMIX2D=0
      IBIHET=0
      DO ISNAP=1,NSNAP
        KPSNAP=LCMGIL(JPSNAP1,ISNAP)
        CALL LCMGET(KPSNAP,'STATE-VECTOR',IPAR)
        NMIX2D=MAX(NMIX2D,IPAR(4))
        IF(NREG2D.NE.IPAR(1)) THEN
          CALL XABORT('SPOASM: INVALID VALUE OF NREG2D.')
        ENDIF
        IF(ISNAP.EQ.1) THEN
          NUNK2D=IPAR(2)
          IBIHET=IPAR(40)
        ELSE
          IF(NUNK2D.NE.IPAR(2)) THEN
            CALL XABORT('SPOASM: INVALID VALUE OF NUNK2D.')
          ELSE IF(IBIHET.NE.IPAR(40)) THEN
            CALL XABORT('SPOASM: INVALID VALUE OF IBIHET.')
          ENDIF
        ENDIF
      ENDDO
      IF(NMIX2D.NE.NMIX) CALL XABORT('SPOASM: INCONSISTENT MIXTURES.')
*----
*  RECOVER MACROLIB INFORMATION
*----
      CALL LCMGET(IPMACR,'STATE-VECTOR',IPAR)
      NGRP=IPAR(1)
      MAXMIX=IPAR(2)
      NANIS=IPAR(3)
      ITRANC=IPAR(6)
      NALBP=IPAR(8)
      NW=IPAR(10)
      IF(NMIX.GT.MAXMIX) THEN
         WRITE(HSMG,'(45HSPOASM: THE NUMBER OF MIXTURES IN THE TRACKIN,
     1   3HG (,I5,49H) IS GREATER THAN THE NUMBER OF MIXTURES IN THE M,
     2   9HACROLIB (,I5,2H).)') NMIX,MAXMIX
         CALL XABORT(HSMG)
      ENDIF
*----
*  CONSTRUCT THE PROPER ORTHOGONAL BASIS
*----
      ALLOCATE(FLUX2D(NUNK2D,NSNAP,NGRP),SOUR2D(NUNK2D,NSNAP,NGRP))
      JPSNAP3=LCMGID(IPSNAP,'FLUX')
      DO ISNAP=1,NSNAP
        KPSNAP=LCMGIL(JPSNAP3,ISNAP)
        IPFLUX=LCMGID(KPSNAP,'FLUX')
        DO IGR=1,NGRP
          CALL LCMGDL(IPFLUX,IGR,FLUX2D(1,ISNAP,IGR))
        ENDDO
        IPSOUR=LCMGID(KPSNAP,'SOUR')
        DO IGR=1,NGRP
          CALL LCMGDL(IPSOUR,IGR,SOUR2D(1,ISNAP,IGR))
        ENDDO
      ENDDO
*----
*   RECOVER RADIAL LEAKAGE CROSS SECTIONS
*----
      ALLOCATE(DB2(NSNAP,NREG2D,NGRP),TXSC(0:MAXMIX,NGRP),
     1 SXSC(0:MAXMIX,NANIS,NGRP))
      DB2(:NSNAP,:NREG2D,:NGRP)=0.D0
      TXSC(0:MAXMIX,:NGRP)=0.0
      SXSC(0:MAXMIX,:NANIS,:NGRP)=0.0
      JPSNAP1=LCMGID(IPSNAP,'TRACK')
      JPSNAP2=LCMGID(IPSNAP,'SYSTEM')
      JPSNAP3=LCMGID(IPSNAP,'FLUX')
      ALLOCATE(LEAK1D(NGRP))
      DO ISNAP=1,NSNAP
        ! recover axial leakage information
        KPSNAP=LCMGIL(JPSNAP3,ISNAP)
        CALL LCMLEN(KPSNAP,'SPOT-LEAK1D',ILONG,ITYLCM)
        IF(ILONG.EQ.NGRP) THEN
          CALL LCMGET(KPSNAP,'SPOT-LEAK1D',LEAK1D)
        ELSE
          LEAK1D(:NGRP)=0.0
        ENDIF
        ! recover other reaction rates
        KPSNAP=LCMGIL(JPSNAP1,ISNAP)
        CALL LCMGET(KPSNAP,'KEYFLX',IDL2D)
        CALL LCMGET(KPSNAP,'MATCOD',MAT2D)
        KPSNAP=LCMGIL(JPSNAP2,ISNAP)
        CALL LCMGET(KPSNAP,'STATE-VECTOR',IPAR)
        IF(NGRP.NE.IPAR(8)) THEN
          CALL XABORT('SPOASM: INVALID VALUE OF NGRP.')
        ELSE IF(NUNK2D.NE.IPAR(9)) THEN
          CALL XABORT('SPOASM: INVALID VALUE OF NUNK2D.')
        ENDIF   
        LPSNAP=LCMGID(KPSNAP,'GROUP')
        DO IGR=1,NGRP
          ALLOCATE(TXSC2D(0:MAXMIX),SXSC2D(0:MAXMIX,NANIS))
          MPSNAP=LCMGIL(LPSNAP,IGR)
          CALL LCMGET(MPSNAP,'DRAGON-TXSC',TXSC2D)
          CALL LCMGET(MPSNAP,'DRAGON-S0XSC',SXSC2D)
          TXSC(0:MAXMIX,IGR)=TXSC2D(0:MAXMIX)
          SXSC(0:MAXMIX,:NANIS,IGR)=SXSC2D(0:MAXMIX,:NANIS)
          SSUM=0.0
          DO I=1,NREG2D
            IBM=MAT2D(I)
            IF(IBM.EQ.0) CYCLE
            IF(IBM.GT.MAXMIX) CALL XABORT('SPOASM: IBM OVERFLOW.')
            IUNK=IDL2D(I)
            IF(IUNK.EQ.0) CYCLE
            DB2(ISNAP,I,IGR)=-TXSC2D(IBM)+SXSC2D(IBM,1)-LEAK1D(IGR)+
     1      SOUR2D(IUNK,ISNAP,IGR)/FLUX2D(IUNK,ISNAP,IGR)
            SSUM=SSUM+DB2(ISNAP,I,IGR)*FLUX2D(IUNK,ISNAP,IGR)
          ENDDO
          IF(IMPX.GT.1) WRITE(IOUT,100) ISNAP,IGR,SSUM
          DEALLOCATE(SXSC2D,TXSC2D)
        ENDDO
      ENDDO
      DEALLOCATE(LEAK1D)
*----
*   CREATE SPOT SYSTEM OBJECT
*----
      ISTATE(:NSTATE)=0
      ISTATE(1)=1
      ISTATE(2)=1
      ISTATE(3)=1
      ISTATE(5)=1
      ISTATE(6)=1
      ISTATE(8)=NGRP
      ISTATE(9)=NUNKN
      ISTATE(10)=MAXMIX
      ISTATE(11)=ISCAT
      ISTATE(14)=-1
      CALL LCMPUT(IPSYS,'STATE-VECTOR',NSTATE,1,ISTATE)
      JPSYS=LCMLID(IPSYS,'GROUP',NGRP)
      ALLOCATE(U(NREG2D,NSNAP),W(NSNAP),V(NSNAP,NSNAP))
      U(:NREG2D,:NSNAP)=0.0D0
      DO IGR=1,NGRP
        KPSYS=LCMDIL(JPSYS,IGR)
        CALL LCMPUT(KPSYS,'DRAGON-TXSC',MAXMIX+1,2,TXSC(0,IGR))
        CALL LCMPUT(KPSYS,'DRAGON-S0XSC',(MAXMIX+1)*NANIS,2,
     1  SXSC(0,1,IGR))
        CALL LCMPUT(KPSYS,'NREG2D',1,1,NREG2D)
        CALL LCMPUT(KPSYS,'NSNAP',1,1,NSNAP)
        CALL LCMPUT(KPSYS,'LEAK2D',NSNAP*NREG2D,4,DB2(1,1,IGR))
      ENDDO
*----
*  SCRATCH STORAGE DEALLOCATION
*----
      DEALLOCATE(IDL2D,MAT2D)
      DEALLOCATE(SXSC,TXSC,DB2,V,W,U,SOUR2D,FLUX2D)
      RETURN
  100 FORMAT(44H SPOASM: RADIAL NEUTRON BALANCE FOR SNAPSHOT,I5,
     1 9H IN GROUP,I5,2H =,1P,E12.4)
      END
