*DECK SPOT
      SUBROUTINE SPOT(NENTRY,HENTRY,IENTRY,JENTRY,KENTRY)
*
*-----------------------------------------------------------------------
*
*Purpose:
* Synthesis Proper Orthogonal Tracking (SPOT) operator.
*
*Copyright:
* Copyright (C) 2024 Ecole Polytechnique de Montreal
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version
*
*Author(s): A. Hebert
*
*Parameters: input/output
* NENTRY  number of LCM objects or files used by the operator.
* HENTRY  name of each LCM object or file:
*         HENTRY(1) create or modification type(L_TRACK);
*         HENTRY(2) read-only type(L_GEOM).
* IENTRY  type of each LCM object or file:
*         =1 LCM memory object; =2 XSM file; =3 sequential binary file;
*         =4 sequential ascii file.
* JENTRY  access of each LCM object or file:
*         =0 the LCM object or file is created;
*         =1 the LCM object or file is open for modifications;
*         =2 the LCM object or file is open in read-only mode.
* KENTRY  LCM object address or file unit number.
*
*-----------------------------------------------------------------------
*
      USE GANLIB
*----
*  SUBROUTINE ARGUMENTS
*----
      INTEGER      NENTRY,IENTRY(NENTRY),JENTRY(NENTRY)
      TYPE(C_PTR)  KENTRY(NENTRY)
      CHARACTER    HENTRY(NENTRY)*12
*----
*  LOCAL VARIABLES
*----
      PARAMETER(NSTATE=40,IOUT=6)
      TYPE(C_PTR) IPTRK,IPSNAP,JPSNAP,KPSNAP,IPGEOM
      INTEGER NCODE(6),ICODE(6)
      REAL ZCODE(6)
      CHARACTER TEXT4*4,TEXT12*12,TITLE*72,HSIGN*12,HSMG*131
      DOUBLE PRECISION DFLOTT
      LOGICAL LBIHET,ILK,LFIXUP
      INTEGER IGP(NSTATE),ISTATE(NSTATE)
      LOGICAL, PARAMETER :: LSHOOT=.FALSE.
      INTEGER, PARAMETER :: EELEM=1,IBFP=0,ESCHM=1
*----
*  ALLOCATABLE ARRAYS
*----
      INTEGER, ALLOCATABLE, DIMENSION(:) :: MAT1D,IDL1D,IDL2D,MAT,IDL,
     1 ISPLX,ISPLY,ISPLZ,IL,IM
      INTEGER, ALLOCATABLE, DIMENSION(:,:) :: MAT2D
      INTEGER, ALLOCATABLE, DIMENSION(:,:,:) :: KEYANI
      REAL, ALLOCATABLE, DIMENSION(:) :: VOL1D,VOL2D,VOL,XXX,YYY,ZZZ,
     1 UU,WW,PL,WX,WE,CST,MN,DN
*----
*  PARAMETER VALIDATION
*----
      IF(NENTRY.LE.2) CALL XABORT('SPOT: THREE PARAMETERS EXPECTED.')
      IF((IENTRY(1).NE.1).AND.(IENTRY(1).NE.2)) CALL XABORT('SPOT: L'
     1 //'CM OBJECT EXPECTED AT LHS.')
      IF((JENTRY(1).NE.0).AND.(JENTRY(1).NE.1)) CALL XABORT('SPOT: E'
     1 //'NTRY IN CREATE OR MODIFICATION MODE EXPECTED.')
      IF((JENTRY(2).NE.2).OR.((IENTRY(2).NE.1).AND.(IENTRY(2).NE.2)))
     1 CALL XABORT('SPOT: ENTRY IN REAL-ONLY MODE EXPECTED AT RHS.')
      IF((JENTRY(3).NE.2).OR.((IENTRY(3).NE.1).AND.(IENTRY(3).NE.2)))
     1 CALL XABORT('SPOT: ENTRY IN READ-ONLY MODE EXPECTED AT RHS.')
      CALL LCMGTC(KENTRY(2),'SIGNATURE',12,HSIGN)
      IF(HSIGN.NE.'L_GEOM') THEN
         TEXT12=HENTRY(2)
         CALL XABORT('SPOT: SIGNATURE OF '//TEXT12//' IS '//HSIGN//
     1   '. L_GEOM EXPECTED.')
      ENDIF
      CALL LCMGTC(KENTRY(3),'SIGNATURE',12,HSIGN)
      IF(HSIGN.NE.'L_ARCHIVE') THEN
         TEXT12=HENTRY(3)
         CALL XABORT('SPOT: SIGNATURE OF '//TEXT12//' IS '//HSIGN//
     1   '. L_ARCHIVE EXPECTED.')
      ENDIF
      IPTRK=KENTRY(1)
      IPGEOM=KENTRY(2)
      IPSNAP=KENTRY(3)
      CALL LCMGET(IPGEOM,'STATE-VECTOR',ISTATE)
      ITYPE=ISTATE(1)
      LX=ISTATE(3)
      MAXPTS=ISTATE(6)
*
      LBIHET=.FALSE.
      IMPX=1
      TITLE=' '
      IF(JENTRY(1).EQ.0) THEN
         HSIGN='L_TRACK'
         CALL LCMPTC(IPTRK,'SIGNATURE',12,HSIGN)
         HSIGN='SPOT'
         CALL LCMPTC(IPTRK,'TRACK-TYPE',12,HSIGN)
         IQUA10=0
         IBIHET=2
         CALL LCMLEN(IPGEOM,'BIHET',ILONG,ITYLCM)
         LBIHET=(ILONG.NE.0)
         IF(LBIHET) IQUA10=5
         IGP(:NSTATE)=0
      ELSE IF(JENTRY(1).EQ.1) THEN
         CALL LCMGTC(IPTRK,'SIGNATURE',12,HSIGN)
         IF(HSIGN.NE.'L_TRACK') THEN
            TEXT12=HENTRY(1)
            CALL XABORT('SPOT: SIGNATURE OF '//TEXT12//' IS '//HSIGN
     1      //'. L_TRACK EXPECTED.')
         ENDIF
         CALL LCMGTC(IPTRK,'TRACK-TYPE',12,HSIGN)
         IF(HSIGN.NE.'SPOT') THEN
            TEXT12=HENTRY(1)
            CALL XABORT('SPOT: TRACK-TYPE OF '//TEXT12//' IS '//HSIGN
     1      //'. SPOT EXPECTED.')
         ENDIF
         CALL LCMGET(IPTRK,'STATE-VECTOR',IGP)
         NREG2D=IGP(5)
         NFLOOR=IGP(6)
         LBIHET=IGP(40).GT.0
         IF(LBIHET) THEN
            CALL LCMSIX(IPTRK,'BIHET',1)
            CALL LCMGET(IPTRK,'PARAM',ISTATE)
            CALL LCMSIX(IPTRK,'BIHET',2)
            IBIHET=ISTATE(6)
            IQUA10=ISTATE(8)
         ELSE
            IBIHET=0
            IQUA10=0
         ENDIF
         CALL LCMLEN(IPTRK,'TITLE',LENGT,ITYLCM)
         IF(LENGT.GT.0) CALL LCMGTC(IPTRK,'TITLE',72,TITLE)
      ENDIF
*----
* READ DATA
*----
      INSB=0
      ISCHM=1
      IELEM=1
      NLF=0
      ISCAT=0
      FRTM=0.05
      SVDEPS=1.0E-5
      LFIXUP=.FALSE.
      IGLK=0
   15 CALL REDGET(INDIC,NITMA,FLOTT,TEXT4,DFLOTT)
   16 IF(INDIC.EQ.10) GO TO 30
      IF(INDIC.NE.3) CALL XABORT('SPOT: CHARACTER DATA EXPECTED.')
      IF(TEXT4.EQ.'EDIT') THEN
         CALL REDGET(INDIC,IMPX,FLOTT,TEXT4,DFLOTT)
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(1).')
      ELSE IF(TEXT4.EQ.'TITL') THEN
         CALL REDGET(INDIC,NITMA,FLOTT,TITLE,DFLOTT)
         IF(INDIC.NE.3) CALL XABORT('SPOT: TITLE EXPECTED.')
      ELSE IF(TEXT4.EQ.'MAXR') THEN
         CALL REDGET(INDIC,MAXPTS,FLOTT,TEXT4,DFLOTT)
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(2).')
      ELSE IF(TEXT4.EQ.'SCHM') THEN
         CALL REDGET(INDIC,ISCHM,FLOTT,TEXT4,DFLOTT)
         IF(ISCHM.EQ.2) IELEM=2
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(3).')
      ELSE IF(TEXT4.EQ.'DIAM') THEN
         CALL REDGET(INDIC,IELEM,FLOTT,TEXT4,DFLOTT)
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(4).')
         IELEM=IELEM+1
      ELSE IF(TEXT4.EQ.'SN') THEN
         CALL REDGET(INDIC,NLF,FLOTT,TEXT4,DFLOTT)
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(5).')
         IF(MOD(NLF,2).EQ.1) CALL XABORT('SPOT: EVEN ORDER EXPECTED.')
         ISCAT=NLF
         IQUAD=2
      ELSE IF(TEXT4.EQ.'SCAT') THEN
         IF(NLF.EQ.0) CALL XABORT('SPOT: DEFINE SN FIRST.')
         CALL REDGET(INDIC,ISCAT,FLOTT,TEXT4,DFLOTT)
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(6).')
         ISCAT=MIN(ISCAT,NLF)
      ELSE IF(TEXT4.EQ.'SVDE') THEN
         CALL REDGET(INDIC,NITMA,SVDEPS,TEXT4,DFLOTT)
         IF(INDIC.NE.2) CALL XABORT('SPOT: REAL DATA EXPECTED(1).')
      ELSE IF(TEXT4.EQ.'NFIX') THEN
         LFIXUP=.TRUE.
         CALL REDGET(INDIC,ICL1,FLOTT,TEXT4,DFLOTT)
      ELSE IF(TEXT4.EQ.'QUAB') THEN
         CALL REDGET(INDIC,IQUA10,FLOTT,TEXT4,DFLOTT)
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(11).')
      ELSE IF(TEXT4.EQ.'SAPO') THEN
         IBIHET=1
      ELSE IF(TEXT4.EQ.'HEBE') THEN
         IBIHET=2
      ELSE IF(TEXT4.EQ.'SLSI') THEN
         IBIHET=3
         CALL REDGET(INDIC,NITMA,FLOTT,TEXT4,DFLOTT)
         IF (INDIC.NE.2) GOTO 16
         FRTM=FLOTT
      ELSE IF(TEXT4.EQ.'SLSS') THEN
         IBIHET=4
         CALL REDGET(INDIC,NITMA,FLOTT,TEXT4,DFLOTT)
         IF (INDIC.NE.2) GOTO 16
         FRTM=FLOTT
      ELSE IF(TEXT4.EQ.'ONEG') THEN
         INSB=0
      ELSE IF(TEXT4.EQ.'ALLG') THEN
         INSB=1
      ELSE IF(TEXT4.EQ.'GQ') THEN
         ! Galerkin quadrature
         CALL REDGET(INDIC,IGLK,FLOTT,TEXT4,DFLOTT)
         IF(INDIC.NE.1) CALL XABORT('SPOT: INTEGER DATA EXPECTED(12).')
      ELSE IF(TEXT4.EQ.';') THEN
         GO TO 30
      ELSE
         CALL XABORT('SPOT: '//TEXT4//' IS AN INVALID KEY WORD.')
      ENDIF
      GO TO 15
*----
* RECOVER AXIAL GEOMETRY INFORMATION
*----
   30 IF(ITYPE.NE.2) CALL XABORT('SPOT: CAR1D GEOMETRY EXPECTED.')
      IF(IMPX.GT.0) WRITE(IOUT,100) TITLE
      ALLOCATE(MAT1D(MAXPTS),IDL1D(MAXPTS))
      ALLOCATE(XXX(MAXPTS+1),YYY(MAXPTS+1),ZZZ(MAXPTS+1))
      ALLOCATE(ISPLX(MAXPTS),ISPLY(MAXPTS),ISPLZ(MAXPTS))
      CALL READ3D(MAXPTS,MAXPTS,MAXPTS,MAXPTS,IPGEOM,IHEX,IR,ILK,SIDE,
     1 XXX,YYY,ZZZ,IMPX,LX,LY,LZ,MAT1D,NEL,NCODE,ICODE,ZCODE,ISPLX,
     2 ISPLY,ISPLZ,ISPLTH,ISPLTL)
      DEALLOCATE(ISPLZ,ISPLY,ISPLX)
      DEALLOCATE(ZZZ,YYY)
      NFLOOR=LX
      CALL LCMPUT(IPTRK,'MAT1D',NFLOOR,1,MAT1D)
*----
* PERFORM AXIAL 1D TRACKING
*----
      ISCAT=MAX(2,ISCAT)
      IF(IGLK.NE.0) THEN
         NSCT=NLF
      ELSE
         NSCT=ISCAT
      ENDIF
      ALLOCATE(VOL1D(NFLOOR),UU(NLF),WW(NLF),PL(ISCAT*NLF))
      ALLOCATE(WX(IELEM+1),WE(EELEM+1),CST(MAX(IELEM,EELEM)))
      ALLOCATE(MN(NSCT*NLF),DN(NLF*NSCT),IL(NSCT),IM(NSCT))
      CALL SNT1DP(IMPX,NFLOOR,IELEM,NCODE,ZCODE,XXX,NLF,ISCAT,UU,WW,
     1 PL,VOL1D,IDL1D,LL4,NUNK1D,LSHOOT,EELEM,WX,WE,CST,IBFP,ISCHM,
     2 ESCHM,IGLK,MN,DN,IL,IM,IQUAD)
      DEALLOCATE(IM,IL,DN,MN)
      DEALLOCATE(CST,WE,WX)
      CALL LCMPUT(IPTRK,'U',NLF,2,UU)
      CALL LCMPUT(IPTRK,'W',NLF,2,WW)
      CALL LCMPUT(IPTRK,'PL',ISCAT*NLF,2,PL)
      DEALLOCATE(PL,WW,UU)
*----
*  RECOVER 2D TRACKING INFORMATION FROM IPSNAP
*----
      CALL LCMGET(IPSNAP,'LISTDIM',NSNAP)
      JPSNAP=LCMGID(IPSNAP,'TRACK')
      NREG2D=0
      NUNK2D=0
      NMIX=0
      IBIHET=0
      DO ISNAP=1,NSNAP
        KPSNAP=LCMGIL(JPSNAP,ISNAP)
        CALL LCMGET(KPSNAP,'STATE-VECTOR',ISTATE)
        NMIX=MAX(NMIX,ISTATE(4))
        IF(ISNAP.EQ.1) THEN
          NREG2D=ISTATE(1)
          NUNK2D=ISTATE(2)
          IBIHET=ISTATE(40)
        ELSE
          IF(NREG2D.NE.ISTATE(1)) THEN
            CALL XABORT('SPOT: INVALID VALUE OF NREG2D.')
          ELSE IF(NUNK2D.NE.ISTATE(2)) THEN
            CALL XABORT('SPOT: INVALID VALUE OF NUNK2D(1).')
          ELSE IF(IBIHET.NE.ISTATE(40)) THEN
            CALL XABORT('SPOT: INVALID VALUE OF IBIHET.')
          ENDIF
        ENDIF
      ENDDO
*----
*  CONSTRUCT 3D GEOMETRY
*----
      NREG=NFLOOR*NREG2D
      NUNKN=NREG2D*(ISCAT*IELEM*NFLOOR+(NFLOOR+1)+NLF)
      LL4=NREG2D*ISCAT*IELEM*NFLOOR
      LL5=NREG2D*(NFLOOR+1)
      ALLOCATE(MAT2D(NREG,NSNAP),IDL2D(NREG),VOL2D(NREG))
      ALLOCATE(MAT(NREG),IDL(NREG),VOL(NREG))
      MAT2D(:NREG,:NSNAP)=0
      JPSNAP=LCMGID(IPSNAP,'TRACK')
      DO ISNAP=1,NSNAP
        KPSNAP=LCMGIL(JPSNAP,ISNAP)        
        CALL LCMGET(KPSNAP,'MATCOD',MAT2D(1,ISNAP))
      ENDDO
      DO IFLOOR=1,NFLOOR
        ISNAP=MAT1D(IFLOOR)
        IF(ISNAP.GT.NSNAP) THEN
          WRITE(HSMG,'(14HSPOT: SNAPSHOT,I4,16H IS NOT DEFINED.)') ISNAP
          CALL XABORT(HSMG)
        ENDIF
        KPSNAP=LCMGIL(JPSNAP,ISNAP)        
        CALL LCMGET(KPSNAP,'KEYFLX',IDL2D)
        CALL LCMGET(KPSNAP,'VOLUME',VOL2D)
        DO J=1,NREG2D
          IOF=(J-1)*NFLOOR+IFLOOR
          VOL(IOF)=VOL2D(J)*(XXX(IFLOOR+1)-XXX(IFLOOR))
          MAT(IOF)=MAT2D(J,ISNAP)
          IDL(IOF)=(IOF-1)*ISCAT*IELEM+1
        ENDDO
      ENDDO
      CALL LCMPUT(IPTRK,'MATCOD',NREG,1,MAT)
      CALL LCMPUT(IPTRK,'KEYFLX',NREG,1,IDL)
      CALL LCMPUT(IPTRK,'VOLUME',NREG,2,VOL)
      CALL LCMPUT(IPTRK,'NCODE',2,1,NCODE)
      CALL LCMPUT(IPTRK,'ICODE',2,1,ICODE)
      CALL LCMPUT(IPTRK,'ZCODE',2,2,ZCODE)
      DEALLOCATE(VOL,MAT)
      DEALLOCATE(VOL2D,IDL2D,MAT2D,VOL1D)
      DEALLOCATE(IDL1D,MAT1D,XXX)
*----
*  PROCESS DOUBLE HETEROGENEITY (BIHET) DATA (IF AVAILABLE)
*----
      IF(LBIHET) CALL XDRTBH(IPGEOM,IPTRK,IQUA10,IBIHET,IMPX,FRTM)
*----
*  SAVE STATE-VECTOR INFORMATION.
*----
      IF(TITLE.NE.' ') CALL LCMPTC(IPTRK,'TITLE',72,TITLE)
      TEXT12=HENTRY(2)
      IGP(1)=NREG
      IGP(2)=NUNKN
      IF(ILK) THEN
         IGP(3)=0
      ELSE
         IGP(3)=1
      ENDIF
      IGP(4)=NMIX
      IGP(6)=NREG2D
      IGP(7)=NFLOOR
      IGP(8)=NSNAP
      IGP(9)=IELEM
      IGP(10)=ISCHM
      IGP(11)=LL4
      IGP(12)=LL5
      IGP(15)=NLF
      IGP(16)=ISCAT
      IF(LFIXUP) IGP(18)=1
      IGP(27)=INSB
      IGP(37)=IGLK
      CALL LCMPUT(IPTRK,'STATE-VECTOR',NSTATE,1,IGP)
      CALL LCMPTC(IPTRK,'LINK.GEOM',12,TEXT12)
      CALL LCMPUT(IPTRK,'EPSI',1,2,EPSI)
      IF(IMPX.GT.1) THEN
         CALL LCMGET(IPTRK,'STATE-VECTOR',IGP)
         WRITE(IOUT,110) IGP(:12),IGP(15:16),IGP(18),IGP(27),IGP(37),
     1   IGP(40)
         CALL LCMLIB(IPSNAP)
      ENDIF
*----
*  SET KEYFLX$ANIS
*----
      ALLOCATE(KEYANI(NREG,IELEM,ISCAT))
      DO IFLOOR=1,NFLOOR
        DO J=1,NREG2D
          IOF=(IFLOOR-1)*NREG2D+J
          IND=IDL(IOF)
          DO IE=1,IELEM
            DO L=1,ISCAT
              IF(IND.EQ.0) THEN
                KEYANI(IOF,IE,L)=0
              ELSE
                KEYANI(IOF,IE,L)=IND+(L-1)*IELEM+IE-1
              ENDIF
            ENDDO
          ENDDO
        ENDDO
      ENDDO
      CALL LCMPUT(IPTRK,'KEYFLX$ANIS',NREG*IELEM*ISCAT,1,KEYANI)
      DEALLOCATE(KEYANI,IDL)
      RETURN
*
  100  FORMAT(/1H1,33H SSSSS  PPPPPP   OOOOO  TTTTTTTT ,95(1H*)/
     1 34H SSSSSSS PPPPPPP OOOOOOO TTTTTTTT ,57(1H*),
     2 38H MULTIGROUP VERSION.  A. HEBERT (2024)/
     3 30H SS   SS PP   PP OO   OO    TT/
     4 30H  SS     PPPPPP  OO   OO    TT/
     5 30H    SS   PP      OO   OO    TT/
     6 30H SS   SS PP      OO   OO    TT/
     7 30H SSSSSSS PP      OOOOOOO    TT/
     8 30H  SSSSS  PP       OOOOO     TT//1X,A72/)
  110 FORMAT(23H TRACKING STATE VECTOR:/
     1 7H NREG  ,I8,25H   (NUMBER OF 3D REGIONS)/
     2 7H NUNKN ,I8,29H   (NUMBER OF 3D FLUX VALUES)/
     3 7H ILK   ,I8,39H   (0=LEAKAGE PRESENT/1=LEAKAGE ABSENT)/
     4 7H NMIX  ,I8,32H   (NUMBER OF MATERIAL MIXTURES)/
     5 7H NSURF ,I8,29H   (NUMBER OF OUTER SURFACES)/
     6 7H NREG2D,I8,25H   (NUMBER OF 2D REGIONS)/
     7 7H NFLOOR,I8,27H   (NUMBER OF AXIAL FLOORS)/
     8 7H NSNAP ,I8,24H   (NUMBER OF SNAPSHOTS)/
     9 7H IELEM ,I8,43H   (1=DIAMOND SCHEME/>1=HIGH SPATIAL ORDER)/
     1 7H ISCHM ,I8,46H   (METHOD OF SPATIAL DISCRETISATION/1=HODD/2=,
     2 3HDG)/
     3 7H LL4   ,I8,38H   (NUMBER OF FLUX UNKNOWNS PER GROUP)/
     4 7H LL5   ,I8,43H   (NUMBER OF INTERFACE CURRENTS PER GROUP)/
     5 7H NLF   ,I8,13H   (SN ORDER)/
     6 7H ISCAT ,I8,47H   (1=ISOTROPIC SOURCE/2=LINEARLY ANISOTROPIC S,
     7 6HOURCE)/
     8 7H IFIX  ,I8,29H   (0/1: NEGATIVE FLUX FIXUP)/
     9 7H INSB  ,I8,36H   (0/1: GROUP VECTORIZATION OFF/ON)/
     1 7H IGLK  ,I8,45H   (0=CLASSICAL SN/>0=GALERKIN QUADRATURE)/
     2 7H IBIHET,I8,47H   (0/1: DOUBLE HETEROGENEITY IS NOT/IS ACTIVE))
      END
